export function createController(options) {
    return new PreviewIframeControllerImpl(options);
}
class PreviewIframeControllerImpl {
    constructor(options) {
        this.options = options;
        this.previewBootstrapped = false;
        this.waitingForBootstrapped = false;
        this.lastMessage = null;
        this.viteError = null;
        this.renderingError = null;
        this.onWindowMessage = (event) => {
            const data = event.data;
            const { listener } = this.options;
            switch (data.kind) {
                case "bootstrapped":
                    this.onBootstrapped();
                    break;
                case "before-render":
                case "action":
                case "log-message":
                    listener(data);
                    break;
                case "renderer-updated":
                    this.clearExpectRenderTimeout();
                    listener({
                        kind: "update",
                        viteError: this.viteError,
                        rendering: {
                            kind: "success",
                            variantKey: data.variantKey,
                            variants: data.variants,
                        },
                    });
                    break;
                case "rendering-error":
                    this.renderingError = data.message;
                    this.clearExpectRenderTimeout();
                    listener({
                        kind: "update",
                        viteError: this.viteError,
                        rendering: {
                            kind: "error",
                            error: this.renderingError,
                        },
                    });
                    break;
                case "vite-logs-error":
                    // We listen to HMR errors to learn about reload errors, for
                    // which there is no Vite error. In other cases, we'd rather
                    // prioritise Vite errors.
                    if (!this.viteError) {
                        this.viteError = {
                            type: "error",
                            err: {
                                message: data.message,
                                stack: "",
                            },
                        };
                        listener({
                            kind: "update",
                            viteError: this.viteError,
                            rendering: null,
                        });
                    }
                    break;
                case "vite-error":
                    this.viteError = data.payload;
                    listener({
                        kind: "update",
                        viteError: this.viteError,
                        rendering: null,
                    });
                    break;
                case "vite-before-update":
                    this.viteError = null;
                    listener({
                        kind: "update",
                        viteError: null,
                        rendering: null,
                    });
                    break;
            }
        };
    }
    start() {
        window.addEventListener("message", this.onWindowMessage);
        this.resetIframe();
    }
    stop() {
        window.removeEventListener("message", this.onWindowMessage);
    }
    showLoading() {
        this.send({
            kind: "show-loading",
        });
    }
    loadComponent(options) {
        this.send({
            kind: "render",
            ...options,
        });
    }
    send(message) {
        var _a;
        this.lastMessage = message;
        if (!this.previewBootstrapped && !this.waitingForBootstrapped) {
            this.resetIframe();
            return;
        }
        const iframeWindow = (_a = this.options.getIframe()) === null || _a === void 0 ? void 0 : _a.contentWindow;
        if (!iframeWindow) {
            return;
        }
        this.renderingError = null;
        this.viteError = null;
        iframeWindow.postMessage(message, document.location.href);
        if (message.kind === "render") {
            this.clearExpectRenderTimeout();
            this.expectRenderTimeout = setTimeout(() => {
                console.warn("Expected render did not occur after 5 seconds. Reloading iframe...");
                this.resetIframe();
            }, 5000);
        }
    }
    resetIframe() {
        const iframe = this.options.getIframe();
        this.previewBootstrapped = false;
        this.renderingError = null;
        if (!iframe) {
            return;
        }
        this.waitingForBootstrapped = true;
        iframe.src = `/preview/?t=${Date.now()}`;
    }
    onBootstrapped() {
        this.previewBootstrapped = true;
        this.waitingForBootstrapped = false;
        if (this.lastMessage) {
            this.send(this.lastMessage);
        }
        this.options.listener({
            kind: "bootstrapped",
        });
    }
    clearExpectRenderTimeout() {
        if (this.expectRenderTimeout) {
            clearTimeout(this.expectRenderTimeout);
            this.expectRenderTimeout = null;
        }
    }
}
