"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Previewer = void 0;
const config_1 = require("@previewjs/config");
const vfs_1 = require("@previewjs/vfs");
const assert_never_1 = __importDefault(require("assert-never"));
const fs_extra_1 = require("fs-extra");
const path_1 = __importDefault(require("path"));
const caching_1 = require("./caching");
const server_1 = require("./server");
const vite_manager_1 = require("./vite/vite-manager");
const POSTCSS_CONFIG_FILE = ["postcss.config.js", ".postcssrc.js"];
const FILES_REQUIRING_RESTART = new Set([
    config_1.PREVIEW_CONFIG_NAME,
    "jsconfig.json",
    "tsconfig.json",
    "package.json",
    "package-lock.json",
    "pnpm-lock.yaml",
    "yarn.lock",
    ...POSTCSS_CONFIG_FILE,
]);
const SHUTDOWN_CHECK_INTERVAL = 3000;
const SHUTDOWN_AFTER_INACTIVITY = 10000;
class Previewer {
    constructor(options) {
        this.options = options;
        this.appServer = null;
        this.viteManager = null;
        this.status = { kind: "stopped" };
        this.shutdownCheckInterval = null;
        this.disposeObserver = null;
        this.config = null;
        this.onFileChangeListener = {
            onChange: (absoluteFilePath, info) => {
                var _a, _b;
                absoluteFilePath = path_1.default.resolve(absoluteFilePath);
                const wrapperPath = this.config
                    ? (_a = this.configWithWrapper(this.config).wrapper) === null || _a === void 0 ? void 0 : _a.path
                    : null;
                if (!info.virtual &&
                    (FILES_REQUIRING_RESTART.has(path_1.default.basename(absoluteFilePath)) ||
                        (wrapperPath &&
                            absoluteFilePath ===
                                path_1.default.resolve(this.options.rootDirPath, wrapperPath)))) {
                    if (this.status.kind === "starting" || this.status.kind === "started") {
                        const port = this.status.port;
                        // Packages were updated. Restart.
                        console.log("New dependencies were detected. Restarting...");
                        this.stop()
                            .then(async () => {
                            await this.start(async () => port);
                        })
                            .catch(console.error);
                    }
                    return;
                }
                if (info.virtual) {
                    (_b = this.viteManager) === null || _b === void 0 ? void 0 : _b.triggerReload(absoluteFilePath);
                }
                else if (this.options.onFileChanged) {
                    this.options.onFileChanged(absoluteFilePath);
                }
            },
        };
        this.transformingReader = (0, vfs_1.createStackedReader)([
            options.reader,
            (0, vfs_1.createFileSystemReader)({
                mapping: {
                    from: options.frameworkPlugin.previewDirPath,
                    to: path_1.default.join(options.rootDirPath, "__previewjs_internal__", "renderer"),
                },
                watch: false,
            }),
            (0, vfs_1.createFileSystemReader)({
                mapping: {
                    from: options.previewDirPath,
                    to: options.rootDirPath,
                },
                watch: false,
            }),
        ]);
    }
    async start(allocatePort) {
        let port;
        const statusBeforeStart = this.status;
        switch (statusBeforeStart.kind) {
            case "starting":
                port = statusBeforeStart.port;
                try {
                    await statusBeforeStart.promise;
                }
                catch (e) {
                    console.error(e);
                    this.status = {
                        kind: "stopped",
                    };
                    await this.startFromStopped(port);
                }
                break;
            case "started":
                if (this.shutdownCheckInterval) {
                    clearInterval(this.shutdownCheckInterval);
                    this.shutdownCheckInterval = null;
                }
                port = statusBeforeStart.port;
                break;
            case "stopping":
                port = statusBeforeStart.port;
                try {
                    await statusBeforeStart.promise;
                }
                catch (e) {
                    console.error(e);
                    this.status = {
                        kind: "stopped",
                    };
                }
                await this.start(async () => port);
                break;
            case "stopped":
                port = await allocatePort();
                await this.startFromStopped(port);
                break;
            default:
                throw (0, assert_never_1.default)(statusBeforeStart);
        }
        return port;
    }
    async startFromStopped(port) {
        this.status = {
            kind: "starting",
            port,
            promise: (async () => {
                // PostCSS requires the current directory to change because it relies
                // on the `import-cwd` package to resolve plugins.
                process.chdir(this.options.rootDirPath);
                this.config = await (0, config_1.readConfig)(this.options.rootDirPath);
                this.appServer = new server_1.Server({
                    middlewares: [
                        ...(this.options.middlewares || []),
                        (req, res, next) => {
                            var _a;
                            (_a = this.viteManager) === null || _a === void 0 ? void 0 : _a.middleware(req, res, next);
                        },
                    ],
                });
                if (this.transformingReader.observe) {
                    this.disposeObserver = await this.transformingReader.observe(this.options.rootDirPath);
                }
                this.transformingReader.listeners.add(this.onFileChangeListener);
                this.viteManager = new vite_manager_1.ViteManager({
                    rootDirPath: this.options.rootDirPath,
                    shadowHtmlFilePath: path_1.default.join(this.options.previewDirPath, "index.html"),
                    reader: this.transformingReader,
                    cacheDir: path_1.default.join((0, caching_1.getCacheDir)(this.options.rootDirPath), "vite"),
                    config: this.configWithWrapper(this.config),
                    logLevel: this.options.logLevel,
                    frameworkPlugin: this.options.frameworkPlugin,
                });
                const server = await this.appServer.start(port);
                await this.viteManager.start(server, port);
                this.status = {
                    kind: "started",
                    port,
                };
            })(),
        };
        await this.status.promise;
    }
    async stop(options = {}) {
        if (this.status.kind !== "started") {
            return;
        }
        if (options.onceUnused) {
            if (!(await this.shutdownCheck())) {
                this.shutdownCheckInterval = setInterval(() => {
                    this.shutdownCheck().catch(console.error);
                }, SHUTDOWN_CHECK_INTERVAL);
            }
        }
        else {
            await this.stopNow();
        }
    }
    configWithWrapper(config) {
        if (!config.wrapper) {
            const defaultWrapperPath = this.options.frameworkPlugin.defaultWrapperPath;
            if ((0, fs_extra_1.pathExistsSync)(path_1.default.join(this.options.rootDirPath, defaultWrapperPath))) {
                config.wrapper = {
                    path: defaultWrapperPath,
                };
            }
        }
        return config;
    }
    async shutdownCheck() {
        var _a;
        const lastPingTimestamp = ((_a = this.viteManager) === null || _a === void 0 ? void 0 : _a.getLastPingTimestamp()) || 0;
        if (lastPingTimestamp + SHUTDOWN_AFTER_INACTIVITY > Date.now()) {
            return false;
        }
        if (this.shutdownCheckInterval) {
            clearInterval(this.shutdownCheckInterval);
            this.shutdownCheckInterval = null;
        }
        await this.stopNow();
        return true;
    }
    async stopNow() {
        if (this.status.kind === "starting") {
            await this.status.promise;
        }
        if (this.status.kind !== "started") {
            return;
        }
        this.transformingReader.listeners.remove(this.onFileChangeListener);
        this.status = {
            kind: "stopping",
            port: this.status.port,
            promise: (async () => {
                if (this.disposeObserver) {
                    await this.disposeObserver();
                    this.disposeObserver = null;
                }
                if (this.viteManager) {
                    await this.viteManager.stop();
                    this.viteManager = null;
                }
                if (this.appServer) {
                    await this.appServer.stop();
                    this.appServer = null;
                }
                this.status = {
                    kind: "stopped",
                };
            })(),
        };
        await this.status.promise;
    }
}
exports.Previewer = Previewer;
